/*
** Module   : BrowseWidget.java
** Abstract : server side BROWSE widget class
**
** Copyright (c) 2005-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 NVS 20051026   @23481 Created initial version.
** 002 SIY 20060307   @25125 Major update for complete BROWSE support.
** 003 SIY 20060319   @25125 Fixed update.
** 004 SIY 20060328   @25276 Fixed reposition code, added a number of 
**                           methods to set various flags and attributes.
** 004 SIY 20060406   @25400 Added setCentered() methods. 
** 005 SIY 20060427   @25816 Added another signature for  setRepositionerRow() method.
** 006 ECF 20060501   @25917 Change to getRowCount. PreselectQueryWrapper
**                           is now obsolete; use new P2JQuery.isPreselect method instead.
** 007 ECF 20060618   @27424 Change to setQuery(). Ensure new query is set to be scrolling.
** 008 GES 20060803   @28366 Some cleanup and refactoring.
** 009 GES 20060804   @28430 Honor new iterate() processing which only
**                           happens at the top of the block body.
** 010 GES 20060808   @28490 Use int instead of long for the backing 
**                           methods that support the server exports.
** 011 ECF 20061019   @30532 Changed reposition(). Removed adjustment code for query off-end.
** 012 ECF 20061023   @30610 Made BrowseWidget tolerant of a query which returns no results. When
**                           a query is set into the widget, its errorIfNull property is set to
**                           false. QueryOffEndException is caught and ignored in finishSetup()
**                           when trying to retrieve the first record.
** 013 ECF 20061103   @30963 Implement RepositionListener interface. This is necessary to receive
**                           notifications of reposition events from the associated query.
**                           Made additional changes to support changes in the P2JQuery interface
**                           and to support the assumption that a new record is fetched when
**                           the associated query is repositioned.
** 014 SIY 20061105   @30983 Finished support for notifications and added getCurrentRow() method.
** 015 GES 20061110   @31077 Fixed abend in getCurrentRow().
** 016 SIY 20061114   @31360 Fixed some repositioning issues. Removed initial query positioning
**                           which did break  browse behavior (initially selected row is
**                           the current row in query).
** 017 ECF 20061129   @31479 Fixed getRowCount(). This was aborting with a QueryOffEndException
**                           in the case of a query which had no results. This exception is now
**                           caught and handled.
** 018 ECF 20061204   @31578 Use new P2JQuery.setBrowsed() method. Also, getCurrentRow() reports
**                           -1 instead of 0 for current row if query is off-end, since 0 is a
**                           valid row number.
** 019 SIY 20070316   @32421 Set column ordinal when it is added to browse.
** 020 SIY 20070507   @33425 Changed handling of record set size request and other 
**                           relevant changes.
** 021 SIY 20070515   @33755 Fixed positioning. 
** 022 SIY 20070605   @33963 Fixed restoring of the position after getRows.
** 023 NVS 20070618   @34083 Fixed ignoreRepo state variable rebalancing.
**                           Due to exceptions in finally{} clauses it could have been left
**                           unbalanced at times, causing legitimate requests to be ignored.
** 024 NVS 20070622   @34206 Fixed regression introduced by #022. reposition() calls throw
**                           QueryOffEndException when the result set is empty in getRows().
** 025 ECF 20070928   @35274 Fixed setCurrentRow(). Was repositioning to
**                           first row after running off-end at last row.
** 026 ECF 20071011   @35446 Further adjustment to setCurrentRow(). Avoid
**                           multiple repositions in case of selecting last browse row.
** 027 ECF 20080208   @37029 Fixed fetchSelectedRow(). This method was  improperly using its
**                           argumentas the row index, rather than as an index within the
**                           list of selected rows.
** 028 GES 20080320   @37581 Added TYPE attribute.
** 029 ECF 20080509   @38246 Added safety code to getRows(). We must first
**                           check if any rows are available before trying to fetch them.
** 030 SVL 20080512   @38277 setScrolling() and setErrorIfNull(false) were
**                           removed from setQuery(), reposition(1) was added to setQuery().
** 031 SVL 20080604   @38536 currentRowImpl() is used instead of currentRow().
** 032 SIY 20080625   @38923 Minor fixes in javadoc and removed unnecessary type cast.
** 033 CA  20080925   @39961 Support for dynamic title.
** 034 ECF 20081106   @40318 Improved getRows(). Use a dynamic list instead of a fixed array,
**                           in case the row count changes mid-query. Break fetch loop on
**                           query off-end condition.
** 035 CA  20081203   @40813 getColumns() must set the ID of the browse column, before sending
**                           it to the client.
** 036 CA  20081210   @40830 Events triggered via APPLY will force the buffer to load the
**                           currently selected row.
** 037 CA  20081210   @40842 Implemented CLOSE QUERY statement. On query close, the browse will
**                           reset its data.
** 038 CA  20081212   @40886 Fixed a regression in setCurrentRow, caused by
**                           #036: rowIndex must always be incremented (so it will be 1-based).
** 039 SIY 20090213   @41306 Added 'force' parameter into refreshStartRow signature.
**                           Added forceRefresh() method.
** 040 SIY 20090706   @43055 Removed apply(). Changes in query behavior made it obsolete.
** 041 CA  20090906   @43831 Fixed FETCH-SELECTED-ROW() - an error message  must be shown if
**                           the received index is unknown. Fixed SET-REPOSITIONED-ROW() - if
**                           the received row number is unknown, it must return false.
** 042 SVL 20091030   @44295 Offend state is preserved while execution of the  getRowCount().
** 043 SVL 20091105   @44316 Into getRowCount() set the query cursor final state to "not between
**                           rows" (if the query allows fetch on reposition).
** 044 SIY 20101122          Minor cleanups.
** 045 SIY 20111101          Updated imports.
** 046 SIY 20111202          Updated imports.
** 047 CA  20121022          Added support for optional query and column list 
**                           clauses.
** 048 CA  20130221          Refactored usage of TYPE attribute.
** 049 CA  20130222          Added NPE protection for cases when a widget is used before is 
**                           attached to a frame (after a CREATE WIDGET statement is used).
** 050 OM  20130320          Added implementation for ResultListHandler interface.
** 051 VMN 20130411          Added boolean parameter modal to recordOrShowError() method.
** 052 OM  20130527          Added int64 support for accessors.
** 053 CA  20130927          Added LegacyResource annotation, to determine resource type.
** 054 CA  20131024          Only QUERY resources allow ResultListHandler API calls - so cast the
**                           query to the ResultListHandler type before calling.
** 055 CA  20131115          Changes related to properly instantiate dynamic resources.
** 056 GES 20140627          Changed color attribute method names to match naming standards.
** 057 CA  20140711          Enhanced font support, for GUI clients.
** 058 VIG 20140810          Added DESELECT-ROWS error processing.
** 059 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 060 SVL 20140909          In getRowCount use query.isNativelyPreselect rather than
**                           query.isPreselect in order to correctly detect adaptive queries as
**                           dynamic.
** 061 CA  20141104          Removed set/getWidth: the BaseEntity's set/getWidthChars APIs need to
**                           be emitted/used.
** 062 SVL 20141102          Added some attributes and functions.
** 063 SVL 20141209          Implemented some functions. Updated error messages in
**                           setNumLockedColumns.
** 063 GES 20141124          Added modified and query attributes.
** 064 SVL 20150212          Modified getRows and getCurrentRow.
** 065 SVL 20150305          Added refresh mode for DELETE-CURRENT-ROW. Fixed case of
**                           unavailable records in getRows.
** 066 SVL 20150331          Implemented DELETE-* functions.
** 067 SVL 20150411          Implemented SELECT-{NEXT|PREV}-ROW, ROW-MARKERS, MAX-DATA-GUESS.
*/

package com.goldencode.p2j.ui;

import java.util.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.persist.event.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.ui.client.*;

/**
 * Server side representation of the Browse widget.
 */
@LegacyResource(resource = LegacyResource.BROWSE)
public class BrowseWidget
extends BaseEntity<BrowseConfig>
implements BrowseElement,
           BrowseInterface,
           QueryAssociable,
           RepositionListener,
           ResultListHandler,
           ScrollbarVerticalElement
{
   /** Default value for the MAX-DATA-GUESS attribute. */
   private static int DATA_GUESS_DEFAULT_VALUE = 100;

   /** Underlying query.  May be null. */
   private P2JQuery query = null;

   /** The registered storage for the current browse title. */
   private CharacterExpression dynTitle = null;
   
   /** List of columns in browse. */
   private ArrayList<BrowseColumnWidget> columns =
      new ArrayList<BrowseColumnWidget>();
   
   /** Widget initialization flag. */
   private boolean initDone = false;
   
   /** Flag to ignore browse-driven reposition events */
   private int ignoreRepo = 0;

   /** Cached value of {@link Browse#ilPageRowsCount}, i.e. real DOWN value. */
   private Integer pageRowsCount = null;

   /**
    * Maximum known row index (1-based). <code>0</code> if there are no rows. <code>null</code>
    * if we haven't tried to fetch any rows with the underlying query yet.
    */
   private Integer maximumKnownRowIndex = null;

   /**
    * <code>true</code> if an actual number of fetched rows is used as MAX-DATA-GUESS.
    * <code>false</code> if the default value is used.
    */
   private boolean realDataGuessUsed = false;

   /**
    * Default constructor.
    */
   public BrowseWidget()
   {
      this(false);
   }
   
   /**
    * Default constructor.
    * 
    * @param    dynamic
    *           Flag indicating if this is a static or dynamic resource.
    */
   public BrowseWidget(boolean dynamic)
   { 
      super(dynamic, new BrowseConfig());
   }

   /**
    * Checks whether the widget is write-protected.
    * 
    * @return   <code>true</code> if the widget is write-protected.
    */
   public logical isReadOnly() 
   {
      return new logical(config.readOnly);
   }

   /**
    * Sets the READ-ONLY attribute of the widget.
    * 
    * @param    r
    *           <code>true</code> if the widget should be write-protected.
    */
   public void setReadOnly(logical r) 
   {
      config.readOnly = r.booleanValue();
      pushScreenDefinition();
   }
   
   /**
    * Get the HELP writable attribute.
    * 
    * @return   The current value of the HELP attribute.
    */
   public character getHelp()
   {
      String help = config.help;
      return (help == null) ? new character() : new character(help);
   }

   /**
    * Queries if the SCREEN-VALUE attribute for this widget was changed.
    * 
    * @return  If <code>true</code>, the widget's screen value was changed.
    */
   public logical isModified()
   {
      // TODO: 4GL docs suggest we need to realize the widget when this is called
      UnimplementedFeature.partial("MODIFIED attribute runtime behavior is incomplete");
      return new logical(config.modified);
   }
   
   /**
    * Defines if the SCREEN-VALUE attribute for this widget has changed.
    * 
    * @param   modified
    *          <code>true</code> to mark the screen value as changed.
    */
   public void setModified(boolean modified)
   {
      // TODO: client side is done yet?
      // TODO: this might need to change the screen-buffer state too
      UnimplementedFeature.partial("MODIFIED attribute runtime behavior is incomplete");
      
      config.modified = modified;
      pushScreenDefinition();
   }
   
   /**
    * Defines if the SCREEN-VALUE attribute for this widget has changed.
    * 
    * @param   modified
    *          <code>true</code> to mark the screen value as changed.
    */
   public void setModified(logical modified)
   {
      if (modified == null || modified.isUnknown())
      {
         UnimplementedFeature.todo("MODIFIED attribute unknown behavior is not known.");
         return;
      }
      
      setModified(modified.booleanValue()); 
   }
   
   /**
    * Get the MULTIPLE attribute.
    * 
    * @return   The value of the MULTIPLE attribute.
    */
   public logical isMultiple()
   {
      return new logical(config.multiple);
   }

   /**
    * Set value of the MULTIPLE attribute.
    * 
    * @param    multiple
    *           New value of the attribute.
    */
   public void setMultiple(boolean multiple)
   {
      config.multiple = multiple;
      pushScreenDefinition();
   }
   
   /**
    * Set value of the MULTIPLE attribute.
    * 
    * @param    multiple
    *           New value of the attribute.
    */
   public void setMultiple(logical multiple)
   {
      if (multiple.isUnknown())
         return;
      
      setMultiple(multiple.booleanValue());
   }

   /**
    * Set value of the SCROLLBAR-VERTICAL attribute.
    * 
    * @param    flag
    *           New value of the attribute.
    */
   public void setScrollbarVertical(boolean flag)
   {
      config.scrollbarVertical = flag;
      pushScreenDefinition();
   }

   /**
    * Set value of the SCROLLBAR-VERTICAL attribute.
    * 
    * @param    flag
    *           New value of the attribute.
    */
   public void setScrollbarVertical(logical flag)
   {
      if (flag == null || flag.isUnknown())
      {
         UnimplementedFeature.todo("SCROLLBAR-VERTICAL attribute unknown behavior is not " +
                                   "known.");
         return;
      }
      
      setScrollbarVertical(flag.booleanValue());
   }

   /**
    * Get value of the SCROLLBAR-VERTICAL attribute.
    *
    * @return  value of the attribute.
    */
   public logical isScrollbarVertical()
   {
      return new logical(config.scrollbarVertical);
   }
   
   /**
    * Set OVERLAY attribute.
    * 
    * @param    overlay
    *           New attribute value.
    */
   public void setOverlay(boolean overlay)
   {
      config.overlay = overlay;
      pushScreenDefinition();
   }
   
   /**
    * Set OVERLAY attribute.
    * 
    * @param    overlay
    *           New attribute value.
    */
   public void setOverlay(logical overlay)
   {
      if (overlay.isUnknown())
         return;
      
      setOverlay(overlay.booleanValue());
   }

   /**
    * Set value of the ROW-MARKERS attribute.
    * 
    * @param    markers
    *           New value of the attribute.
    */
   public void setRowMarkers(boolean markers)
   {
      config.rowMarkers = markers;
      pushScreenDefinition();
   }
   
   /**
    * Set value of the ROW-MARKERS attribute.
    * 
    * @param    markers
    *           New value of the attribute.
    */
   public void setRowMarkers(logical markers)
   {
      if (markers.isUnknown())
         return;
      
      setRowMarkers(markers.booleanValue());
   }

   /**
    * Set value of the ROW-MARKERS attribute as opposite to passed value.
    * 
    * @param    nomarkers
    *           New value of the attribute.
    */
   public void setNoRowMarkers(boolean nomarkers)
   {
      setRowMarkers(!nomarkers);
   }
   
   /**
    * Set value of the ROW-MARKERS attribute as opposite to passed value.
    * 
    * @param    nomarkers
    *           New value of the attribute.
    */
   public void setNoRowMarkers(logical nomarkers)
   {
      if (nomarkers.isUnknown())
         return;
      
      setNoRowMarkers(nomarkers.booleanValue());
   }
   
   /**
    * Set SEPARATORS attribute as opposite to passed value.
    * 
    * @param    noseparators
    *           New value for the attribute.
    */
   public void setNoSeparators(boolean noseparators)
   {
      config.separators = !noseparators;
      pushScreenDefinition();
   }

   /**
    * Set SEPARATORS attribute as opposite to passed value.
    * 
    * @param    noseparators
    *           New value for the attribute.
    */
   public void setNoSeparators(logical noseparators)
   {
      if (noseparators.isUnknown())
         return;
      
      setNoSeparators(noseparators.getValue());
   }
   
   /**
    * Set SEPARATORS attribute.
    * 
    * @param    separators
    *           New value for the attribute.
    */
   public void setSeparators(boolean separators)
   {
      config.separators = separators;
      pushScreenDefinition();
   }

   /**
    * Set SEPARATORS attribute.
    * 
    * @param    separators
    *           New value for the attribute.
    */
   public void setSeparators(logical separators)
   {
      if (separators.isUnknown())
         return;
      
      setSeparators(separators.booleanValue());
   }

   /**
    * Set LABELS attribute.
    * 
    * @param    labels
    *           New value for the attribute.
    */
   public void setLabels(boolean labels)
   {
      config.labels = labels;
      pushScreenDefinition();
   }

   /**
    * Set LABELS attribute.
    * 
    * @param    labels
    *           New value for the attribute.
    */
   public void setLabels(logical labels)
   {
      if (labels.isUnknown())
         return;
      
      setLabels(labels.booleanValue());
   }

   /**
    * Set LABELS attribute as opposite to passed value.
    * 
    * @param    nolabels
    *           New value for the attribute.
    */
   public void setNoLabels(boolean nolabels)
   {
      setLabels(!nolabels);
   }

   /**
    * Set CENTERED attribute value.
    * 
    * @param    centered
    *           New value for the attribute.
    */
   public void setCentered(boolean centered)
   {
      config.centered = centered;
      pushScreenDefinition();
   }

   /**
    * Set CENTERED attribute value.
    * 
    * @param    centered
    *           New value for the attribute.
    */
   public void setCentered(logical centered)
   {
      if (centered.isUnknown())
         return;
      
      setCentered(centered.booleanValue());
   }
   
   /**
    * Set LABELS attribute as opposite to passed value.
    * 
    * @param    labels
    *           New value for the attribute.
    */
   public void setNoLabels(logical labels)
   {
      if (labels.isUnknown())
         return;
      
      setNoLabels(labels.booleanValue());
   }

   /**
    * Set BOX attribute.
    * 
    * @param    box
    *           New value for the attribute.
    */
   public void setBox(boolean box)
   {
      config.box = box;
      pushScreenDefinition();
   }

   /**
    * Set BOX attribute.
    * 
    * @param    box
    *           New value for the attribute.
    */
   public void setBox(logical box)
   {
      if (box.isUnknown())
         return;
      
      setBox(box.booleanValue());
   }

   /**
    * Set BOX attribute as opposite to passed value.
    * 
    * @param    nobox
    *           New value for the attribute.
    */
   public void setNoBox(boolean nobox)
   {
      setBox(!nobox);
   }

   /**
    * Set BOX attribute as opposite to passed value.
    * 
    * @param    nobox
    *           New value for the attribute.
    */
   public void setNoBox(logical nobox)
   {
      if (nobox.isUnknown())
         return;
      
      setNoBox(nobox.booleanValue());
   }

   /**
    * Set value of the MULTIPLE attribute as opposite to passed argument.
    * 
    * @param    single
    *           New value of the attribute.
    */
   public void setSingle(logical single)
   {
      if (single.isUnknown())
         return;
      
      setSingle(single.booleanValue());
   }

   /**
    * Set value of the MULTIPLE attribute as opposite to passed argument.
    * 
    * @param    single
    *           New value of the attribute.
    */
   public void setSingle(boolean single)
   {
      setMultiple(!single);
   }

   /**
    * Sets the dynamic title for this browse widget.
    *
    * @param    title
    *           The new value for the dynamic title.
    */
   public void setDynamicTitle(BaseDataType title)
   {
      setDynamicTitle(new AccessorWrapper(title));
   }
   
   /**
    * Sets the dynamic title for this browse widget.
    *
    * @param    title
    *           The new value for the dynamic title.
    */
   public void setDynamicTitle(Accessor title) 
   {
      class TitleExpr
      extends CharacterExpression
      {
         private final Accessor titleExpr;
         
         /**
          * Construct an instance for given {@link Accessor} instance.
          * 
          * @param    titleExpr
          *           Dynamic title variable.
          */
         public TitleExpr(Accessor titleExpr)
         {
            this.titleExpr = titleExpr;
         }
         
         public character execute()
         {
            return (character) titleExpr.get();
         }
      }
      setDynamicTitle(new TitleExpr(title));
   }

   /**
    * Sets the dynamic title for this browse widget.
    *
    * @param    titleExp
    *           The new expression for the dynamic title.
    */
   public void setDynamicTitle(CharacterExpression titleExp) 
   {
      this.dynTitle = titleExp;

      // the current value computed from the dynamic title is used until 
      // the buffer is sent to the client and the title is updated
      character title = dynTitle.execute();
      String    txt   = (title.isUnknown() ? "" : title.toStringMessage());
      config.title = txt;
   }

   /**
    * Set value of the NO-ASSIGN attribute.
    *
    * @param    noAssign
    *           New value of the attribute.
    */
   public void setNoAssign(boolean noAssign)
   {
      config.noAssign = noAssign;
      pushScreenDefinition();
   }

   /**
    * Set value of the NO-ASSIGN attribute.
    *
    * @param    noAssign
    *           New value of the attribute.
    */
   public void setNoAssign(logical noAssign)
   {
      if (noAssign.isUnknown())
         return;

      setNoAssign(noAssign.getValue());
   }

   /**
    * Set value of the NO-UNDERLINE attribute.
    *
    * @param    noUnderline
    *           New value of the attribute.
    */
   public void setNoUnderline(boolean noUnderline)
   {
      config.noUnderline = noUnderline;
      pushScreenDefinition();
   }

   /**
    * Set value of the NO-UNDERLINE attribute.
    *
    * @param    noUnderline
    *           New value of the attribute.
    */
   public void setNoUnderline(logical noUnderline)
   {
      if (noUnderline.isUnknown())
         return;

      setNoUnderline(noUnderline.getValue());
   }

   /**
    * Get value of the FIT-LAST-COLUMN attribute.
    *
    * @return  value of the attribute.
    */
   public logical isFitLastColumn()
   {
      return new logical(config.fitLastColumn);
   }

   /**
    * Set value of the FIT-LAST-COLUMN attribute.
    *
    * @param    fitLastColumn
    *           New value of the attribute.
    */
   public void setFitLastColumn(boolean fitLastColumn)
   {
      config.fitLastColumn = fitLastColumn;
      pushScreenDefinition();
   }

   /**
    * Set value of the FIT-LAST-COLUMN attribute.
    *
    * @param    fitLastColumn
    *           New value of the attribute.
    */
   public void setFitLastColumn(logical fitLastColumn)
   {
      if (fitLastColumn.isUnknown())
         return;

      setFitLastColumn(fitLastColumn.getValue());
   }

   /**
    * Set value of the EXPANDABLE attribute.
    *
    * @param    expandable
    *           New value of the attribute.
    */
   public void setExpandable(boolean expandable)
   {
      setFitLastColumn(expandable);
   }

   /**
    * Set value of the EXPANDABLE attribute.
    *
    * @param    expandable
    *           New value of the attribute.
    */
   public void setExpandable(logical expandable)
   {
      if (expandable.isUnknown())
         return;

      setExpandable(expandable.getValue());
   }

   /**
    * Get value of the EXPANDABLE attribute.
    *
    * @return    value of the attribute.
    */
   public logical isExpandable()
   {
      return isFitLastColumn();
   }

   /**
    * Set value of the COLUMN-RESIZABLE attribute.
    *
    * @param    resizable
    *           <code>true</code> you can resize a browse column by pointing, clicking, and
    *           dragging.
    */
   public void setColumnResizable(boolean resizable)
   {
      setColumnResizable(new logical(resizable));
   }

   /**
    * Set value of the COLUMN-RESIZABLE attribute.
    *
    * @param    resizable
    *           <code>true</code> you can resize a browse column by pointing, clicking, and
    *           dragging.
    */
   public void setColumnResizable(logical resizable)
   {
      if (resizable.isUnknown())
         return;

      config.columnResizable = resizable.booleanValue();
      pushScreenDefinition();
   }

   /**
    * Get value of the COLUMN-RESIZABLE attribute.
    *
    * @return   <code>true</code> you can resize a browse column by pointing, clicking, and
    *           dragging.
    */
   public logical isColumnResizable()
   {
      return new logical(config.columnResizable);
   }

   /**
    * Set value of the COLUMN-MOVABLE attribute.
    *
    * @param    movable
    *           <code>true</code> you can move a browse column by pointing, clicking, and
    *           dragging.
    */
   public void setColumnMovable(boolean movable)
   {
      setColumnMovable(new logical(movable));
   }

   /**
    * Set value of the COLUMN-MOVABLE attribute.
    *
    * @param    movable
    *           <code>true</code> you can move a browse column by pointing, clicking, and
    *           dragging.
    */
   public void setColumnMovable(logical movable)
   {
      if (movable.isUnknown())
         return;

      config.columnMovable = movable.booleanValue();
      pushScreenDefinition();
   }

   /**
    * Get value of the CREATE-ON-ADD attribute.
    *
    * @return   value of the attribute.
    */
   public logical isCreateOnAdd()
   {
      return new logical(config.createOnAdd);
   }

   /**
    * Set value of the CREATE-ON-ADD attribute.
    *
    * @param    createOnAdd
    *           New value of the attribute.
    */
   public void setCreateOnAdd(boolean createOnAdd)
   {
      setCreateOnAdd(new logical(createOnAdd));
   }

   /**
    * Set value of the CREATE-ON-ADD attribute.
    *
    * @param    createOnAdd
    *           New value of the attribute.
    */
   public void setCreateOnAdd(logical createOnAdd)
   {
      if (createOnAdd.isUnknown())
         return;

      config.createOnAdd = createOnAdd.booleanValue();
      pushScreenDefinition();
   }

   /**
    * Get value of the COLUMN-MOVABLE attribute.
    *
    * @return   <code>true</code> you can move a browse column by pointing, clicking, and
    *           dragging.
    */
   public logical isColumnMovable()
   {
      return new logical(config.columnMovable);
   }

   /**
    * Set value of the REFRESHABLE attribute.
    *
    * @param    refreshable
    *           <code>true</code> if the rows that appear in a browse are refreshed when an
    *           application opens or repositions a query.
    */
   public void setRefreshable(boolean refreshable)
   {
      setRefreshable(new logical(refreshable));
   }

   /**
    * Set value of the REFRESHABLE attribute.
    *
    * @param    refreshable
    *           <code>true</code> if the rows that appear in a browse are refreshed when an
    *           application opens or repositions a query.
    */
   public void setRefreshable(logical refreshable)
   {
      if (refreshable.isUnknown())
         return;

      config.refreshable = refreshable.booleanValue();
      pushScreenDefinition();
   }

   /**
    * Get value of the REFRESHABLE attribute.
    *
    * @return   <code>true</code> if the rows that appear in a browse are refreshed when an
    *           application opens or repositions a query.
    */
   public logical isRefreshable()
   {
      return new logical(config.refreshable);
   }

   /**
    * Get value of the CURRENT-ROW-MODIFIED attribute.
    *
    * @return   <code>true</code> if any cells in the current row have been changed.
    */
   public logical isCurrentRowModified()
   {
      //TODO implement
      return new logical();
   }

   /**
    * Get value of the NEW-ROW attribute.
    *
    * @return   <code>true</code> if the row in focus was added to the browse using the
    *           INSERT-ROW() method and has not been added to the database.
    */
   public logical isNewRow()
   {
      //TODO implement
      return new logical();
   }

   /**
    * Get value of the ROW-MARKERS attribute.
    *
    * @return   <code>true</code> if the browse has row markers.
    */
   public logical hasRowMarkers()
   {
      if (!config.rowMarkers)
         return new logical(false);

      boolean editable = false;
      for (BrowseColumnWidget column : columns)
      {
         if (!column.isReadOnly().booleanValue())
         {
            editable = true;
            break;
         }
      }

      return new logical(editable);
   }

   /**
    * Gets the LABEL-BGCOLOR writable attribute.
    *
    * @return   The current value of the LABEL-BGCOLOR attribute.
    */
   public integer getLabelBgColor()
   {
      // TODO implement. Looks like it is accessor for LABEL-BGCOLOR attribute of one of the
      // columns.
      return new integer();
   }

   /**
    * Sets the LABEL-BGCOLOR writable attribute.
    *
    * @param    bgcolor
    *           The new value for the LABEL-BGCOLOR attribute.
    */
   public void setLabelBgColor(integer bgcolor)
   {
      // TODO implement. Looks like it is accessor for LABEL-BGCOLOR attribute of one of the
      // columns.
   }

   /**
    * Sets the LABEL-BGCOLOR writable attribute.
    *
    * @param    bgcolor
    *           The new value for the LABEL-BGCOLOR attribute.
    */
   public void setLabelBgColor(int bgcolor)
   {
      // TODO implement. Looks like it is accessor for LABEL-BGCOLOR attribute of one of the
      // columns.
   }

   /**
    * Set value of the COLUMN-SCROLLING attribute.
    *
    * @param    columnScrolling
    *           New value of the attribute.
    */
   public void setColumnScrolling(boolean columnScrolling)
   {
      config.columnScrolling = columnScrolling;
      pushScreenDefinition();
   }

   /**
    * Set value of the COLUMN-SCROLLING attribute.
    *
    * @param    columnScrolling
    *           New value of the attribute.
    */
   public void setColumnScrolling(logical columnScrolling)
   {
      if (columnScrolling.isUnknown())
         return;

      setColumnScrolling(columnScrolling.getValue());
   }

   /**
    * Set value of the COLUMN-SCROLLING attribute as opposite to passed argument.
    *
    * @param    noColumnScrolling
    *           New value of the attribute.
    */
   public void setNoColumnScrolling(boolean noColumnScrolling)
   {
      config.columnScrolling = !noColumnScrolling;
      pushScreenDefinition();
   }

   /**
    * Set value of the COLUMN-SCROLLING attribute as opposite to passed argument.
    *
    * @param    noColumnScrolling
    *           New value of the attribute.
    */
   public void setNoColumnScrolling(logical noColumnScrolling)
   {
      if (noColumnScrolling.isUnknown())
         return;

      setNoColumnScrolling(noColumnScrolling.getValue());
   }

   /**
    * Set value of the ROW-HEIGHT-CHARS attribute.
    *
    * @param    rowHeightChars
    *           New value of the attribute.
    */
   public void setRowHeightChars(double rowHeightChars)
   {
      config.rowHeightChars = rowHeightChars;
      pushScreenDefinition();
   }

   /**
    * Set value of the ROW-HEIGHT-CHARS attribute.
    *
    * @param    rowHeightChars
    *           New value of the attribute.
    */
   public void setRowHeightChars(NumberType rowHeightChars)
   {
      if (rowHeightChars.isUnknown())
         return;

      setRowHeightChars(rowHeightChars.doubleValue());
   }

   /**
    * Sets the TITLE writable attribute.
    *
    * @param    title
    *           The new value for the TITLE attribute.
    */
   public void setTitle(String title)
   {
      super.setTitle(title);
      
      // if the browse TITLE attribute is set, then the dynamic title
      // must be canceled
      this.dynTitle = null;
   }

   /**
    * Get value of NUM-COLUMNS attribute.
    * 
    * @return   Number of columns in the browse.
    */
   public integer getNumColumns()
   {
      return new integer(columns.size());
   }

   /**
    * Get specified browse column.
    * 
    * @param    index
    *           1-based column index.
    *           
    * @return   reference to column widget.
    */
   public BrowseColumnWidget getBrowseColumn(double index)
   {
      BrowseColumnWidget res = columns.get((int) index - 1);
      
      if (res == null)
      {
         ErrorManager.recordOrShowError(4065, 
             "**The GET-BROWSE-COLUMN attribute on the BROWSE " + 
             frame.getName(this) + " has invalid arguments.",
             true);
      }
      
      return res;
   }

   /**
    * Get specified browse column.
    * 
    * @param    index
    *           1-based column index.
    *           
    * @return   reference to column widget.
    */
   public BrowseColumnWidget getBrowseColumn(NumberType index)
   {
      double ndx = (index.isUnknown()) ? 0 : index.doubleValue();
      
      return getBrowseColumn(ndx);
   }
   
   /**
    * Get the NUM-SELECTED-ROWS attribute.
    * 
    * @return   The current value of NUM-SELECTED-ROWS attribute.
    */
   public integer getNumSelectedRows()
   {
      return new integer(LogicalTerminal.getNumSelectedRows(this));
   }

   /**
    * Get the QUERY writable attribute.
    * 
    * @return   Current value of the QUERY attribute.
    */
   public P2JQuery getQuery()
   {
      return query;
   }

   /**
    * Set the query which will back this browse.  Ensure it is scrollable and
    * does not raise an error condition if it goes off-end.
    * 
    * @param    query
    *           New value for the attribute.
    */
   public void setQuery(P2JQuery query)
   {
      this.query = query;
      query.setBrowsed(true);
      query.setFetchOnReposition(true);
      query.addRepositionListener(this);
      query.reposition(1);
      checkMaxDataGuess();
   }
   
   /**
    * Get the query associated with the given resource.
    *
    * @return   The associated query or <code>unknown</code> if none exists.
    */
   public handle getQueryAsHandle()
   {
      P2JQuery query = getQuery();
      
      return (query == null) ? new handle() : new handle((QueryWrapper) query);
   }
   
   /**
    * Set the query associated with the given resource.
    *
    * @param    qry
    *           The query to associate.
    */
   public void setQueryAsHandle(handle qry)
   {
      if (qry == null || qry.isUnknown() || !qry._isValid())
      {
         UnimplementedFeature.todo("QUERY attribute setter unknown behavior not investigated.");
         return;
      }
      
      Object resource = qry.get();
      
      if (!(resource instanceof QueryWrapper))
      {
         UnimplementedFeature.todo("QUERY attribute setter wrong resource behavior not investigated.");
         return;
      }
      
      setQuery((QueryWrapper) resource);
   }

   /**
    * Implements the SELECT-ALL() widget method.
    * 
    * @return   Always <code>true</code>.
    */
   public logical selectAll()
   {
      if (!isMultiple().booleanValue())
      {
         final String msg = "You cannot set SELECT-ALL for a single select BROWSE brws";
         ErrorManager.recordOrShowError(4674, msg, false, false);
         return new logical(false);
      }

      LogicalTerminal.selectAll(this);
      return new logical(true);
   }

   /**
    * Implements the DESELECT-ROWS() widget method.
    * 
    * @return   Always <code>true</code>.
    */
   public logical deselectRows()
   {
      if (!checkNoRecords("DESELECT-ROWS"))
      {
         return new logical(false);
      }

      if (frame == null) 
      {
         String[] msg = new String[]
         { "Unable to realise BROWSE widget because it is not in a frame",
           "Unknown error code 1 for attribute ? on the BROWSE widget",
            "**Unable to query attribute DESELECT-ROWS for BROWSE widget" };
         int[] num = new int[] {4040, 4104, 4077};
         ErrorManager.recordOrShowError(num, msg, false);
         return new logical(false);
      }
      
      LogicalTerminal.deselectRows(this);
      return new logical(true);
   }

   /**
    * Implements the FETCH-SELECTED-ROW() widget method.
    * 
    * @param   n
    *          1-based index of the selected row to fetch;  note that this is
    *          the index into the list of selected rows within the browse,
    *          NOT the index of the row itself.
    *
    * @return  <code>true</code> if method succeeds.
    */
   public logical fetchSelectedRow(int n)
   {
      return fetchSelectedRow(new int64(n));
   }
   
   /**
    * Implements the FETCH-SELECTED-ROW() widget method.
    * 
    * @param   n
    *          1-based index of the selected row to fetch;  note that this is
    *          the index into the list of selected rows within the browse,
    *          NOT the index of the row itself.
    *
    * @return  <code>true</code> if method succeeds.
    */
   public logical fetchSelectedRow(int64 n)
   {
      if (query == null)
      {
         final String msg = "Browse widget must contain at least one record" + 
                            " before FETCH-SELECTED-ROW method is used";
         ErrorManager.displayError(2108, msg, false);

         return new logical(false);
      }

      if (n == null || n.isUnknown())
      {
         final String msg = "Invalid row index specified for browse method " + 
                            "FETCH-SELECTED-ROW";
         ErrorManager.displayError(385, msg, false);

         return new logical(false);
      }
      
      int selListIndex = n.intValue() - 1;
      if (selListIndex < 0)
         return new logical(false);
      
      // Get the 0-based row index of the (n - 1)th selected row.
      int uiRow = LogicalTerminal.getSelectedRowIndex(this, selListIndex);
      if (uiRow < 0)
         return new logical(false);
      
      // Reposition the query to this row and update the underlying buffers.
      reposition(new integer(uiRow + 1), true);
      
      if (query.isOffEnd().booleanValue())
         return new logical(false);
      
      return new logical(true);
   }

   /**
    * Implements the REFRESH() widget method.
    *
    * @return  Always <code>true</code>.
    */
   public logical refresh()
   {
      return refresh(false);
   }

   /**
    * Refresh the browse cache and redisplay the browse.
    *
    * @param   deleteCurrentRowMode
    *          <code>true</code> if this function is used for update on DELETE-CURRENT-ROW.
    *          <code>false</code> if this function is used for update on REFRESH.
    *
    * @return  Always <code>true</code>.
    */
   public logical refresh(boolean deleteCurrentRowMode)
   {
      LogicalTerminal.refresh(this, deleteCurrentRowMode);
      return new logical(true);
   }

   /**
    * Implements the SELECT-FOCUSED-ROW() widget method.<p>
    * Selects the row that currently has focus in a browse widget, even if it is not currently
    * displayed.
    * 
    * @return   Always <code>true</code>.
    */
   public logical selectFocusedRow()
   {
      LogicalTerminal.selectFocusedRow(this);
      return new logical(true);
   }

   /**
    * Implements the DESELECT-FOCUSED-ROW() widget method.<p>
    * Deselects the row with current focus.
    *
    * @return   <code>true</code> if the browse has at least one record.
    */
   public logical deselectFocusedRow()
   {
      if (!checkNoRecords("DESELECT-FOCUSED-ROW"))
      {
         return new logical(false);
      }

      LogicalTerminal.deselectFocusedRow(this);
      return new logical(true);
   }

   /**
    * Implements the SELECT-ROW() widget method.<p>
    * Selects the specified row if it is currently in the browse viewport. In a single-select
    * browse, the previously selected row is deselected. No rows are deselected in a
    * multiple-select browse.
    *
    * @param   row
    *          1-based row index in the browse viewport.
    *
    * @return  <code>true</code> if the row with the given index exist in the browse viewport.
    */
   public logical selectRow(integer row)
   {
      if (!checkNoRecords("SELECT-ROW"))
      {
         return new logical(false);
      }

      Integer val = row.isUnknown() ? null : row.intValue();
      if (val == null ||
          val < 1     ||
          val > getBrowsePageRowCount() ||
          !LogicalTerminal.selectRow(this, val))
      {
         errorInvalidRowIndex("SELECT-ROW");
         return new logical(false);
      }

      return new logical(true);
   }

   /**
    * Implements the SELECT-ROW() widget method.<p>
    * Selects the specified row if it is currently in the browse viewport. In a single-select
    * browse, the previously selected row is deselected. No rows are deselected in a
    * multiple-select browse.
    *
    * @param   row
    *          1-based row index in the browse viewport.
    *
    * @return  <code>true</code> if the row with the given index exist in the browse viewport.
    */
   public logical selectRow(int row)
   {
      return selectRow(new integer(row));
   }

   /**
    * Implements the SELECT-NEXT-ROW() widget method.<p>
    * Deselects all currently selected rows in a browse and selects the row after the deselected
    * row.<p>
    * This method also repositions the query to the new row and copies the record into the
    * database buffer. The browse automatically scrolls to the selected row if it is out of
    * view.<p>
    * This method is intended for use with a browse that supports the selection of a single row at
    * a time (MULTIPLE attribute is set to FALSE). If more than one row is selected when you
    * execute this method, all of the selected rows are deselected and the record after the last
    * selected row becomes the selected row.
    *
    * @return  <code>true</code> on success.
    */
   public logical selectNextRow()
   {
      final String methodName = "SELECT-NEXT-ROW";
      if (!checkNoRecords(methodName))
      {
         return new logical(false);
      }

      try
      {
         return LogicalTerminal.selectNextRow(this);
      }
      catch (ErrorConditionException e)
      {
         handleError(e, methodName);
         return new logical(); //correct
      }
   }

   /**
    * Implements the SELECT-PREV-ROW() widget method.<p>
    * Deselects all currently selected rows in a browse and selects the row before the deselected
    * row.<p>
    * This method also repositions the query to the new row and copies the record into the
    * database buffer.The browse automatically scrolls to the selected row if it is out of
    * view.<p>
    * This method is intended for use with a browse that supports the selection of a single row at
    * a time (MULTIPLE attribute is set to FALSE). If more than one row is selected when you
    * execute this method, all of the selected rows are deselected and the record before the last
    * selected row becomes the selected row.
    *
    * @return  <code>true</code> on success.
    */
   public logical selectPreviousRow()
   {
      final String methodName = "SELECT-PREV-ROW";
      if (!checkNoRecords(methodName))
      {
         return new logical(false);
      }

      try
      {
         return LogicalTerminal.selectPreviousRow(this);
      }
      catch (ErrorConditionException e)
      {
         handleError(e, methodName);
         return new logical(); //correct
      }
   }

   /**
    * Implements the DESELECT-SELECTED-ROW() widget method.<p>
    * Deselects the <i>n</i>th selected row in a browse.<p>
    * Progress maintains a numbered list of selected rows, starting at 1. When the
    * DESELECT-SELECTED-ROW(n) method is encountered, Progress searches this list to find
    * the <i>n</i>th selected row.
    *
    * @param   row
    *          1-based index of a selected row (rows are numbered in the order they were
    *          selected).
    *
    * @return  <code>true</code> if the row is successfully deselected.
    */
   public logical deselectSelectedRow(integer row)
   {
      if (!checkNoRecords("DESELECT-SELECTED-ROW"))
      {
         return new logical(false);
      }

      Integer val = row.isUnknown() ? null : row.intValue();
      if (val == null || val < 1 || !LogicalTerminal.deselectSelectedRow(this, val))
      {
         errorInvalidRowIndex("DESELECT-SELECTED-ROW");
         return new logical(false);
      }

      return new logical(true);
   }

   /**
    * Implements the DESELECT-SELECTED-ROW() widget method.<p>
    * Deselects the <i>n</i>th selected row in a browse.<p>
    * Progress maintains a numbered list of selected rows, starting at 1. When the
    * DESELECT-SELECTED-ROW(n) method is encountered, Progress searches this list to find
    * the <i>n</i>th selected row.
    *
    * @param   row
    *          1-based index of a selected row (rows are numbered in the order they were
    *          selected).
    *
    * @return  <code>true</code> if the row is successfully deselected.
    */
   public logical deselectSelectedRow(int row)
   {
      return deselectSelectedRow(new integer(row));
   }

   /**
    * Implements the DELETE-SELECTED-ROW() widget method.<p>
    * Deletes the <i>n</i>th selected row from a browse and the results list.<p>
    * Progress maintains a numbered list of selected rows, starting at 1. When the
    * DELETE-SELECTED-ROW(n) method is encountered, Progress searches this list to find
    * the <i>n</i>th selected row.<p>
    * This method does not delete the record from the database and has no effect on the database
    * buffer.
    *
    * @param   row
    *          1-based index of a row to delete (rows are numbered in the order they were
    *          selected).
    *
    * @return  <code>true</code> if the row is successfully deleted.
    */
   public logical deleteSelectedRow(integer row)
   {
      final String methodName = "DELETE-SELECTED-ROW";
      if (row.isUnknown())
      {
         errorInvalidRowIndex(methodName);
         return new logical(false);
      }

      return deleteSingleRow(row.intValue() - 1, methodName);
   }

   /**
    * Implements the DELETE-SELECTED-ROW() widget method.<p>
    * Deletes the <i>n</i>th selected row from a browse and the results list.<p>
    * Progress maintains a numbered list of selected rows, starting at 1. When the
    * DELETE-SELECTED-ROW(n) method is encountered, Progress searches this list to find
    * the <i>n</i>th selected row.<p>
    * This method does not delete the record from the database and has no effect on the database
    * buffer.
    *
    * @param   row
    *          1-based index of a row to delete (rows are numbered in the order they were
    *          selected).
    *
    * @return  <code>true</code> if the row is successfully deleted.
    */
   public logical deleteSelectedRow(int row)
   {
      return deleteSelectedRow(new integer(row));
   }

   /**
    * Implements the DELETE-SELECTED-ROWS() widget method.<p>
    * Deletes all currently selected rows from a browse and the associated results list.<p>
    * This method does not delete the record from the database and has no effect on the database
    * buffer.
    *
    * @return  <code>true</code> if rows were successfully deleted.
    */
   public logical deleteSelectedRows()
   {
      final String methodName = "DELETE-SELECTED-ROWS";

      int[] selectedRows;
      try
      {
         selectedRows = LogicalTerminal.getSelectedRowIndexes(this);
      }
      catch (ErrorConditionException e)
      {
         handleError(e, methodName);
         return new logical(false);
      }

      Arrays.sort(selectedRows);

      for (int i = selectedRows.length - 1; i >= 0; i--)
      {
         logical res = deleteRow(selectedRows[i], false);
         if (!res.booleanValue())
            return res;
      }

      LogicalTerminal.refreshAfterDeletion(this, -1);
      return new logical(true);
   }

   /**
    * Implements the DELETE-CURRENT-ROW() widget method.<p>
    * Deletes the most recently selected row from a browse and the results list.<p>
    * This method does not delete the record from the database and has no effect on the database
    * buffer.
    *
    * @return  <code>true</code> if the row is successfully deleted.
    */
   public logical deleteCurrentRow()
   {
      return deleteSingleRow(-1, "DELETE-CURRENT-ROW");
   }

   /**
    * Implements the IS-ROW-SELECTED() widget method.<p>
    * Returns <code>true</code> if a specified row in the browse viewport is currently selected.
    *
    * @param   row
    *          1-based row index in the browse viewport.
    *
    * @return  <code>true</code> if a specified row in the browse viewport is currently selected.
    */
   public logical isRowSelected(integer row)
   {
      if (!checkNoRecords("IS-ROW-SELECTED"))
      {
         return new logical(false);
      }

      Integer val = row.isUnknown() ? null : row.intValue();
      if (val == null                   ||
          val < 1                       ||
          val > getBrowsePageRowCount() ||
          val > query.getNumResults().intValue())
      {
         errorInvalidRowIndex("IS-ROW-SELECTED");
         return new logical(false);
      }

      boolean res = LogicalTerminal.isRowSelected(this, val);
      return new logical(res);
   }

   /**
    * Implements the IS-ROW-SELECTED() widget method.<p>
    * Returns <code>true</code> if a specified row in the browse viewport is currently selected.
    *
    * @param   row
    *          1-based row index in the browse viewport.
    *
    * @return  <code>true</code> if a specified row in the browse viewport is currently selected.
    */
   public logical isRowSelected(int row)
   {
      return isRowSelected(new integer(row));
   }


   /**
    * Implements the INSERT-ROW() widget method.<p>
    * Inserts a blank line in an updateable browse before or after the last selected row. The
    * blank line is a placeholder for a new record to be added through the browse. This method
    * cannot be used with a read-only browse.
    *
    * @param   mode
    *          <code>"BEFORE"</code> if a row should be added before the current browse row.
    *          <code>"AFTER"</code> if a row should be added after the current browse row.
    *
    * @return  <code>true</code> if a row was successfully added.
    */
   public logical insertRow(character mode)
   {
      // TODO implement
      return new logical(false);
   }

   /**
    * Implements the INSERT-ROW() widget method.<p>
    * Inserts a blank line in an updateable browse before or after the last selected row. The
    * blank line is a placeholder for a new record to be added through the browse. This method
    * cannot be used with a read-only browse.
    *
    * @param   mode
    *          <code>"BEFORE"</code> if a row should be added before the current browse row.
    *          <code>"AFTER"</code> if a row should be added after the current browse row.
    *
    * @return  <code>true</code> if a row was successfully added.
    */
   public logical insertRow(String mode)
   {
      return insertRow(new character(mode));
   }

   /**
    * Implements the INSERT-ROW() widget method.<p>
    * Inserts a blank line in an updateable browse before the last selected row. The
    * blank line is a placeholder for a new record to be added through the browse. This method
    * cannot be used with a read-only browse.
    *
    * @return  <code>true</code> if a row was successfully added.
    */
   public logical insertRow()
   {
      return insertRow(new character("BEFORE"));
   }

   /**
    * Implements the ADD-LIKE-COLUMN() widget method.<p>
    * Creates a browse column from the specified field and returns its widget handle.
    *
    * @param  fieldName
    *         The name of a field in one of the buffers associated with the browse’s query. If the
    *         query is a join, the name must be qualified with the database name.
    * @param  position
    *         1-based position of the browse column.
    *
    * @return handle to the created column.
    */
   public handle addLikeColumn(character fieldName, integer position)
   {
      //TODO implement
      return new handle();
   }

   /**
    * Implements the ADD-LIKE-COLUMN() widget method.<p>
    * Creates a browse column from the specified field and returns its widget handle.
    *
    * @param  fieldName
    *         The name of a field in one of the buffers associated with the browse’s query. If the
    *         query is a join, the name must be qualified with the database name.
    * @param  position
    *         1-based position of the browse column.
    *
    * @return handle to the created column.
    */
   public handle addLikeColumn(String fieldName, int position)
   {
      return addLikeColumn(new character(fieldName), new integer(position));
   }

   /**
    * Implements the ADD-LIKE-COLUMN() widget method.<p>
    * Creates a browse column from the specified field and returns its widget handle.
    *
    * @param  fieldHandle
    *         Handle to a field in one of the buffers associated with the browse’s query.
    * @param  position
    *         1-based position of the browse column.
    *
    * @return handle to the created column.
    */
   public handle addLikeColumn(handle fieldHandle, integer position)
   {
      //TODO implement
      return new handle();
   }

   /**
    * Implements the ADD-LIKE-COLUMN() widget method.<p>
    * Creates a browse column from the specified field and returns its widget handle.
    *
    * @param  fieldHandle
    *         Handle to a field in one of the buffers associated with the browse’s query.
    * @param  position
    *         1-based position of the browse column.
    *
    * @return handle to the created column.
    */
   public handle addLikeColumn(handle fieldHandle, int position)
   {
      return addLikeColumn(fieldHandle, new integer(position));
   }

   /**
    * Implements the ADD-LIKE-COLUMN() widget method.<p>
    * Creates a browse column from the specified field, adds it to the end of the columns and
    * returns its widget handle.
    *
    * @param  fieldName
    *         The name of a field in one of the buffers associated with the browse’s query. If the
    *         query is a join, the name must be qualified with the database name.
    *
    * @return handle to the created column.
    */
   public handle addLikeColumn(character fieldName)
   {
      return addLikeColumn(fieldName, new integer(0));
   }

   /**
    * Implements the ADD-LIKE-COLUMN() widget method.<p>
    * Creates a browse column from the specified field, adds it to the end of the columns and
    * returns its widget handle.
    *
    * @param  fieldName
    *         The name of a field in one of the buffers associated with the browse’s query. If the
    *         query is a join, the name must be qualified with the database name.
    *
    * @return handle to the created column.
    */
   public handle addLikeColumn(String fieldName)
   {
      return addLikeColumn(new character(fieldName));
   }

   /**
    * Implements the ADD-LIKE-COLUMN() widget method.<p>
    * Creates a browse column from the specified field, adds it to the end of the columns and
    * returns its widget handle.
    *
    * @param  fieldHandle
    *         Handle to a field in one of the buffers associated with the browse’s query.
    *
    * @return handle to the created column.
    */
   public handle addLikeColumn(handle fieldHandle)
   {
      return addLikeColumn(fieldHandle, new integer(0));
   }

   /**
    * Implements the MOVE-COLUMN() widget method.<p>
    * Repositions a column in a browse widget.
    *
    * @param  src
    *         1-based position of the column to be moved.
    * @param  dest
    *         1-based position to which the column is moved.
    *
    * @return <code>true</code> if the column was successfully moved.
    */
   public logical moveColumn(integer src, integer dest)
   {
      //TODO implement
      return new logical(false);
   }

   /**
    * Implements the MOVE-COLUMN() widget method.<p>
    * Repositions a column in a browse widget.
    *
    * @param  src
    *         1-based position of the column to be moved.
    * @param  dest
    *         1-based position to which the column is moved.
    *
    * @return <code>true</code> if the column was successfully moved.
    */
   public logical moveColumn(int src, int dest)
   {
      return moveColumn(new integer(src), new integer(dest));
   }

   /**
    * Implements the GET-REPOSITIONED-ROW() widget method.<p>
    * Returns the row index of the browse viewport where the REPOSITION TO ROWID (or
    * RECID) statement displays a repositioned record.
    *
    * @return the row index of the browse viewport where the REPOSITION TO ROWID (or
    *         RECID) statement displays a repositioned record
    */
   public integer getRepositionedRow()
   {
      //TODO implement
      return new integer();
   }

   /**
    * Implements the FIRST-COLUMN attribute getter.<p>
    *
    * @return a handle to the first column in a browse widget, regardless of the value of its
    *         READ-ONLY attribute or its VISIBLE attribute.
    */
   public handle getFirstColumn()
   {
      if (getNumColumns().intValue() == 0)
         return new handle();

      return getBrowseColumn(1).asWidgetHandle();
   }

   /**
    * Implements the FOCUSED-ROW attribute getter.<p>
    *
    * @return the 1-based index of the focused row in the viewport.
    */
   public integer getFocusedRow()
   {
      return new integer(LogicalTerminal.getFocusedRow(this));
   }

   /**
    * Implements the NUM-LOCKED-COLUMNS attribute getter.<p>
    *
    * @return the number of visible leading columns locked in a browse widget.
    */
   public integer getNumLockedColumns()
   {
      return new integer(config.numLockedColumns);
   }

   /**
    * Implements the NUM-LOCKED-COLUMNS attribute setter.<p>
    *
    * @param num
    *        The number of visible leading columns locked in a browse widget.
    */
   public void setNumLockedColumns(double num)
   {
      setNumLockedColumns(new decimal(num));
   }

   /**
    * Implements the NUM-LOCKED-COLUMNS attribute setter.<p>
    *
    * @param num
    *        The number of visible leading columns locked in a browse widget.
    */
   public void setNumLockedColumns(NumberType num)
   {
      if (num.isUnknown())
      {
         ErrorManager.displayError(4083,
               "**Unable to assign UNKNOWN value to attribute NUM-LOCKED-COLUMNS on BROWSE " +
                     name().toStringMessage(),
               false);
         return;
      }

      int val = num.intValue();

      if (val < 0)
      {
         ErrorManager.displayError(4082,
               "**Attribute NUM-LOCKED-COLUMNS must be greater than or equal to 0 on BROWSE " +
                     name().toStringMessage(),
               false);
         return;
      }

      if (val > columns.size())
      {
         ErrorManager.displayError(4533,
               "You cannot set NUM-LOCKED-COLUMNS to be greater than the total number of " +
                  "columns for BROWSE " + name().toStringMessage(),
               false);
         ErrorManager.recordOrThrowError(3131,
               "Unable to set attribute NUM-LOCKED-COLUMNS in widget " +
                  name().toStringMessage() + " of type BROWSE",
               false);
         return;
      }

      if (initDone)
      {
         //TODO separate ChUI and GUI cases
         // In ChUI you cannot change the value of the attribute after browse was displayed.
         ErrorManager.displayError(4088,
               "**Progress does not support the NUM-LOCKED-COLUMNS attribute for the BROWSE " +
                  name().toStringMessage() + " in this display environment",
               false);
         return;
      }

      config.numLockedColumns = val;
      pushScreenDefinition();
   }

   /**
    * Implements the MIN-HEIGHT-CHARS attribute getter.<p>
    *
    * @return  minimum height of the browse, in character units.
    */
   public decimal getMinHeightChars()
   {
      // TODO seems to be a calculated attribute
      return new decimal();
   }

   /**
    * Implements the ROW-HEIGHT-CHARS attribute getter.<p>
    *
    * @return row height, in characters, of the browse.
    */
   public decimal getRowHeightChars()
   {
      return new decimal(config.rowHeightChars);
   }

   /**
    * Implements the ALLOW-COLUMN-SEARCHING attribute getter.<p>
    *
    * @return <code>true</code> if column searching is allowed for the browse.
    */
   public logical isAllowColumnSearching()
   {
      return new logical(config.allowColumnSearching);
   }

   /**
    * Implements the ALLOW-COLUMN-SEARCHING attribute setter.<p>
    *
    * @param allow
    *        <code>true</code> if column searching is allowed for the browse.
    */
   public void setAllowColumnSearching(boolean allow)
   {
      setAllowColumnSearching(new logical(allow));
   }

   /**
    * Implements the ALLOW-COLUMN-SEARCHING attribute setter.<p>
    *
    * @param allow
    *        <code>true</code> if column searching is allowed for the browse.
    */
   public void setAllowColumnSearching(logical allow)
   {
      if (allow.isUnknown())
         return;

      config.allowColumnSearching = allow.booleanValue();
      pushScreenDefinition();
   }

   /**
    * Implements the SEPARATORS attribute getter.<p>
    *
    * @return <code>true</code> if row and column separators of the browse widget are displayed.
    */
   public logical isSeparators()
   {
      return new logical(config.separators);
   }

   /**
    * Set value of the MAX-DATA-GUESS attribute.
    *
    * @param    maxDataGuess
    *           The estimated number of records in a browse query.
    */
   public void setMaxDataGuess(double maxDataGuess)
   {
      setMaxDataGuess(new decimal(maxDataGuess));
   }

   /**
    * Set value of the MAX-DATA-GUESS attribute.
    *
    * @param    maxDataGuess
    *           The estimated number of records in a browse query.
    */
   public void setMaxDataGuess(NumberType maxDataGuess)
   {
      if (maxDataGuess.isUnknown())
         return;

      config.maxDataGuess = maxDataGuess.intValue();
      pushScreenDefinition();
   }

   /**
    * Get value of the MAX-DATA-GUESS attribute.
    *
    * @return   the estimated number of records in a browse query.
    */
   public integer getMaxDataGuess()
   {
      return new integer(config.maxDataGuess);
   }

   /**
    * Implements the SET-REPOSITIONED-ROW() widget method.
    * 
    * @param    row
    *           1-based row number where the new record is displayed
    * @param    cond
    *           if <code>true</code>, tries to reposition to the browse
    *           viewport existing row first
    * 
    * @return   Always <code>true</code>.
    */
   public logical setRepositionedRow(int row, boolean cond)
   {
      LogicalTerminal.setRepositionedRow(this, row - 1, cond);
      return new logical(true);
   }
   
   /**
    * Provides the SET-REPOSITIONED-ROW() method which assigns the row number and type of
    * repositioning for the browse.
    * 
    * @param    row
    *           1-based row number where the new record is displayed
    * @param    cond
    *           If this parameter is "conditional", then method tries to
    *           reposition to the browse viewport existing row first
    * 
    * @return   The result of operation. It evaluates to <code>true</code> if
    *           operation was successful.
    */
   public logical setRepositionedRow(int64 row, String cond)
   {
      if (row == null || row.isUnknown())
      {
         return new logical(false);
      }
      
      return setRepositionedRow(row.intValue(), cond);
   }
   
   /**
    * Provides the SET-REPOSITIONED-ROW() method which assigns the row number and type of
    * repositioning for the browse.
    * 
    * @param    row
    *           1-based row number where the new record is displayed
    * @param    cond
    *           If this parameter is "conditional", then method tries to
    *           reposition to the browse viewport existing row first
    * 
    * @return   The result of operation. It evaluates to <code>true</code> if
    *           operation was successful.
    */
   public logical setRepositionedRow(int row, String cond)
   {
      if (cond == null)
      {
         return new logical(false);
      }
      
      return setRepositionedRow(row, "conditional".startsWith(cond.toLowerCase()));
   }
   
   /**
    * Blocks attempts to set the DCOLOR attribute as it is read-only.
    * 
    * @param   d
    *          Flag is not used.
    */
   public void setDColor(int64 d)
   {
      throw new RuntimeException("DCOLOR attribute is read only");
   }

   /**
    * Blocks attempts to get the PFCOLOR attribute as it is undefined.
    * 
    * @return  nothing 
    */
   public integer getPfColor()
   {
      throw new RuntimeException("PFCOLOR attribute is undefined");
   }

   /**
    * Blocks attempts to set the PFCOLOR attribute as it is undefined.
    * 
    * @param   p
    *          Flag is not used.
    */
   public void setPfColor(int64 p)
   {
      throw new RuntimeException("PFCOLOR attribute is undefined");
   }

   /**
    * Set the TITLE-FONT attribute of this browse, dialog, frame or pop-up menu.
    * 
    * @param    fontNum
    *           An entry in the font-table or unknown to refer the default font.
    */
   @Override
   public void setTitleFont(int64 fontNum)
   {
      // if the font is unknown, it gets set to UNDEFINED.
      int font = (fontNum.isUnknown() ? FontManager.UNDEFINED_FONT : fontNum.intValue());

      config.titleFont = font;
      pushScreenDefinition();
   }

   /**
    * Get the TITLE-FONT attribute of this browse, dialog, frame or pop-up menu.
    * 
    * @return   A value representing an entry in the user's font-table.
    */
   @Override
   public integer getTitleFont()
   {
      int font = config.titleFont;
      
      return (font == FontManager.UNDEFINED_FONT ? new integer() : new integer(font));
   }

   /**
    * This will creates an entry in the result list for the current row.  This implements the
    * 4GL CREATE-RESULT-LIST-ENTRY() method.
    * 
    * @return  true on success.
    */
   @Override
   public logical createResultListEntry()
   {
      return query != null ? ((ResultListHandler) query).createResultListEntry()
                           : new logical(false);
   }

   /**
    * This will delete the current row in the query's results.  This implements the
    * 4GL DELETE-RESULT-LIST-ENTRY() method.
    *
    * @return  true on success.
    */
   @Override
   public logical deleteResultListEntry()
   {
      return deleteResultListEntry(false, true);
   }

   /**
    * This will delete the current row in the query's results.  This implements the
    * 4GL DELETE-RESULT-LIST-ENTRY() method.
    *
    * @param  allowBetweenRows
    *         If <code>true</code> then the cursor can be positioned between rows (the next
    *         row will be deleted). <code>false</code> for conventional DELETE-RESULT-LIST-ENTRY()
    *         mode where the cursor should be positioned on a row (otherwise <code>false</code> is
    *         returned).
    * @param  refreshBrowse
    *         If <code>true</code> then the browse is refreshed.
    * 
    * @return  true on success.
    */
   public logical deleteResultListEntry(boolean allowBetweenRows, boolean refreshBrowse)
   {
      logical res = query != null ? query.deleteResultListEntry(allowBetweenRows)
                                  : new logical(false);
      if (res.booleanValue())
      {
         if (refreshBrowse)
            refresh(true);

         checkMaxDataGuess();
         if (maximumKnownRowIndex > 0)
            maximumKnownRowIndex--;

         if (realDataGuessUsed)
            setMaxDataGuess(maximumKnownRowIndex);
      }

      return res;
   }

   // ------------------------------------------------------------------
   // Client side BrowseModel request handlers
   // ------------------------------------------------------------------

   /**
    * Return information about columns.
    * 
    * @return   Meta information about columns.
    */
   public BrowseColumnConfig[] getColumns()
   {
      BrowseColumnConfig[] res = new BrowseColumnConfig[columns.size()];
      int i = 0;
      
      ConfigManager mgr = ConfigManager.getInstance();
      
      for (BrowseColumnWidget column : columns)
      {
         BrowseColumnConfig cfg = mgr.duplicate(column.config);
         cfg.syncWithWidget();

         cfg.id = new WidgetId(column.getId());
         
         res[i++] = cfg;
      }
      
      return res;
   }

   /**
    * Get number of available rows starting from the given one.
    * TODO May only be used for preselect queries in GUI. Do not use before adapting it to that
    * use case.
    *
    * @param   startRow
    *          Initial row to start count rows from.
    *
    * @return  Number of available rows or 0 if no query specified.
    */
   public int getRowCount(int startRow)
   {
      if (query == null)
      {
         return 0;
      }

      int res = 0;

      integer rowId = query.currentRowImpl();
      boolean isFoR = query.isFetchOnReposition();
      OffEnd offEnd = query.getOffEnd();

      try
      {
         ignoreRepo++;

         query.setFetchOnReposition(false);

         if (query.isNativelyPreselect())
         {
            query.first();
            res = query.size().intValue();
         }
         else
         {
            query.forward(Browse.CACHE_SIZE * 5);
            res = getCurrentRow();
            if (res < 0)
               res = 0;
         }
      }
      catch (QueryOffEndException exc)
      {
         // Thrown if query.first() cannot move to a first result because the
         // result set is empty.
      }
      finally
      {
         ignoreRepo--;
         query.setFetchOnReposition(isFoR);

         if (!rowId.isUnknown())
         {
            reposition(rowId, true);
         }
         else if (offEnd == OffEnd.BACK)
         {
            query.last();
         }
         else if (res > 0)
         {
            query.first();
         }
      }

      return res;
   }

   /**
    * Retrieves several subsequent rows from the database.
    * 
    * @param    startRowIndex
    *           Index of the first row to retrieve.
    * @param    count
    *           Maximum count of rows to return.
    * @param    forward
    *           <code>true</code> if rows are retrieved in the forward direction,
    *           <code>false</code> if rows are retrieved in the backward direction,
    * 
    * @return   Array of rows. Each row is array of cell values, but the first element of the row
    *           array is 0-based index of the row in the browse model.
    *           Note that if rows are fetched in the backward direction, the resulting rows are
    *           returned in the straight order starting from the row with the least row number.
    *           If no query specified, an array of size 0 is returned.
    */
   public BaseDataType[][] getRows(int startRowIndex, int count, boolean forward)
   {
      if (query == null || !query.isOpen().booleanValue())
      {
         return new BaseDataType[0][0];
      }

      ArrayList<BaseDataType[]> data = new ArrayList<BaseDataType[]>(count);
      
      if (count == 0)
      {
         return (new BaseDataType[0][columns.size() + 1]);
      }

      integer rowId = query.currentRowImpl();
      boolean fOr = query.isFetchOnReposition();
      int numBuffers = query.numBuffers().intValue();

      try
      {
         ignoreRepo++;

         integer repositionRow = new integer(startRowIndex + (forward ? 1 : 2));
         reposition(repositionRow, false);
         query.setFetchOnReposition(true);
         checkMaxDataGuess();

         rowLoop:
         for (int i = 0; i < count; i++)
         {
            if (forward)
               query.next();
            else
               query.previous();

            if (query._isOffEnd())
            {
               break;
            }

            // In some cases query may not be off-end but there is no current available record
            // (e.g. after DELETE-CURRENT-ROW).
            for (int bufferIndex = 1; bufferIndex <= numBuffers; bufferIndex++)
            {
               handle bufferHandle = query.bufferHandle(bufferIndex);
               if (!bufferHandle.unwrapBuffer()._available())
               {
                  break rowLoop;
               }
            }

            BaseDataType[] cols = new BaseDataType[columns.size() + 1];
            Iterator<BrowseColumnWidget> iter = columns.iterator();
            for (int j = 1; iter.hasNext(); j++)
            {
               BrowseColumnWidget column = iter.next();
               BaseDataType val = column.getValue();
               
               cols[j] = val;
            }

            integer row = query.currentRowImpl();
            if (row.isUnknown())
            {
               StringBuilder errorString = new StringBuilder("Cannot get number of the row [");
               for (int j = 1; j < cols.length; j++)
               {
                  if (j != 1)
                     errorString.append(", ");
                  errorString.append(cols[j].toStringExport());
               }
               errorString.append("]");

               throw new IllegalStateException(errorString.toString());
            }

            if (row.intValue() > maximumKnownRowIndex)
            {
               maximumKnownRowIndex = row.intValue();
            }

            row.decrement();
            cols[0] = row;

            data.add(cols);
         }

         // update maxDataGuess
         if (forward)
         {
            if (maximumKnownRowIndex > getMaxDataGuess().intValue() ||
                data.size() < count)
            {
               // We have exceeded current data guess or we have reached the last row
               setMaxDataGuess(maximumKnownRowIndex);
               realDataGuessUsed = true;
            }
         }
      }
      catch (QueryOffEndException exc)
      {
         // potentially, may be thrown in reposition() because the result set
         // is empty.
      }
      finally
      {
         ignoreRepo--;

         try
         {
            // Now reposition to where we were originally.
            if (!rowId.isUnknown())
               reposition(rowId, true);
            else
               query.first();
         }
         catch (QueryOffEndException exc)
         {
            // thrown in reposition() if the result set is empty.
         }
         finally
         {
            query.setFetchOnReposition(fOr);
         }
      }

      if (!forward)
      {
         Collections.reverse(data);
      }
      BaseDataType[][] result = data.toArray(new BaseDataType[data.size()][]);
      return result;
   }
   
   /**
    * Updates selected cells in the given row or nothing to do if no query
    * specified.
    * 
    * @param    rowIndex
    *           0-based index of the row to update.
    * @param    cellIndices
    *           Cells for update indexes.
    * @param    cellValues
    *           Cells for update new values.
    * 
    * @throws   DataValidationException
    *           New values did not pass database validation rules.
    */
   public void updateRow(int            rowIndex,
                         int[]          cellIndices,
                         BaseDataType[] cellValues)
   throws DataValidationException
   {
      if (query == null)
      {
         return;
      }

      integer rowId = query.currentRowImpl();
      String msg = null;
      
      reposition(new integer(rowIndex + 1), true);

      BrowseColumnWidget[] c = new BrowseColumnWidget[columns.size()];
      
      c = columns.toArray(c);
      
      TransactionManager.pushScope("update",
                                   TransactionManager.TRANSACTION,
                                   false,
                                   false,
                                   false,
                                   false);
      try
      {
         update:
         {
            TransactionManager.blockSetup();
            
            do
            {
               try
               {
                  for (int i = 0; i < cellIndices.length; i++)
                  {
                     c[cellIndices[i]].setValue(cellValues[i]);
                  }
               }
               catch (ErrorConditionException err_except)
               {
                  TransactionManager.honorError(err_except);
                  TransactionManager.rollback("update");
                  TransactionManager.triggerRetry(null);
                  msg = err_except.getMessage();
               }
               catch (RetryUnwindException ru_except)
               {
                  TransactionManager.honorRetry(ru_except);
               }
            }
            while (TransactionManager.needsRetry());
            
            if (TransactionManager.isBreakPending())
            {
               break update;
            }
         }
      }
      finally
      {
         TransactionManager.popScope();
      }      
      
      reposition(rowId, true);
      
      if (msg != null)
         throw new DataValidationException(msg);
   }

   /**
    * Returns <code>true</code> if specified row is new and is not present
    * in the data base.
    * 
    * @param    rowIndex
    *           Index of the row to return information about.
    *          
    * @return   <code>true</code> if row is new and <code>false</code>
    *           otherwise.
    */
   public boolean isNewRow(int rowIndex)
   {
      if (query == null)
      {
         return false;
      }

      integer save = query.currentRowImpl();
      
      reposition(new integer(rowIndex + 1), false);

      boolean res = query.isOffEnd().booleanValue(); 
      
      reposition(save, false);

      return res;
   }

   /**
    * Set new current row in the browse. This method is called as notification
    * from the client that current row is changed.
    * 
    * @param   rowIndex
    *          0-based index of the current row.
    * @param   force
    *          Force positioning to the specified row, even if it is selected.          
    *          
    * @return  New current row or 0 if no query specified.
    */
   public int setCurrentRow(int rowIndex, boolean force)
   {
      if (query == null || !query.isOpen().booleanValue() || rowIndex < 0)
      {
         return 0;
      }

      if (!force && rowIndex == getCurrentRow())
      {
         return rowIndex;
      }
      
      ignoreRepo++;
      boolean fOr = query.isFetchOnReposition();
      
      try
      {
         query.setFetchOnReposition(true);
         reposition(new integer(rowIndex + 1), true);
         logical offEnd = query.isOffEnd();
         if (!offEnd.isUnknown() && offEnd.booleanValue())
         {
            // If we've gone too far, reset ourselves to the last record.
            query.last();
         }
      }
      catch (QueryOffEndException exc)
      {
         // Thrown if result set is empty.
      }
      finally
      {
         ignoreRepo--;
         query.setFetchOnReposition(fOr);
      }
      
      return getCurrentRow();
   }

   /**
    * Get current row in the query.
    *
    * @return   The current row index (0-based) or -1 if the query is off-end.
    */
   public int getCurrentRow()
   {
      if (query == null)
      {
         return -1;
      }

      integer row = query.currentRow();
      return (row.isUnknown() ? -1 : row.intValue() - 1);
   }
   
   //---------------------------------------------------------------
   
   /**
    * Set new value for the DOWN attribute.
    * 
    * @param    down
    *           New value for the attribute.
    */
   public void setDown(int down)
   {
      config.down = down;
      pushScreenDefinition();
   }

   /**
    * Add new field to the list of browse fields.
    * 
    * @param    column
    *           Field to add.
    */
   public void addColumn(BrowseColumnWidget column)
   {
      ((BrowseColumnConfig) column.getConfig()).ordinal = columns.size();
      columns.add(column);
      column.setBrowseId(getId());
   } 

   /**
    * Change editable state for all columns.
    * 
    * @param    editable
    *           New value for the attribute.
    */
   public void setEditableAll(boolean editable)
   {
      for (BrowseColumnWidget column : columns)
      {
         column.setEditable(editable);
      }
   }

   /**
    * Gets the SCROLLABLE writeable attribute.
    *
    * @return   The current value of the SCROLLABLE attribute.
    */
   public logical isScrollable()
   {
      return new logical(config.scrollable);
   }

   /**
    * Sets the SCROLLABLE writeable attribute.
    *
    * @param    scroll
    *           The new value for the SCROLLABLE attribute.
    */
   public void setScrollable(logical scroll)
   {
      if (scroll.isUnknown())
         return;

      setScrollable(scroll.booleanValue());
   }

   /**
    * Sets the SCROLLABLE writeable attribute.
    *
    * @param    scroll
    *           The new value for the SCROLLABLE attribute.
    */
   public void setScrollable(boolean scroll)
   {
      config.scrollable = scroll;
      pushScreenDefinition();
   }

   /**
    * Sets the BGCOLOR writable attribute.
    *
    * @param    bgcolor
    *           The new value for the BGCOLOR attribute.
    */
   @Override
   public void setBgColor(integer bgcolor)
   {
      int color = (bgcolor == null || bgcolor.isUnknown()) ? -1 : bgcolor.intValue();
      setBgColor(color);
   }

   /**
    * Register query and list of fields with the browse.
    * 
    * @param    query
    *           Reference to query. May be null.
    * 
    * @param    list
    *           List of fields. May be null.
    */
   public void registerQuery(P2JQuery query, FrameElement[] list)
   {
      if (query != null)
      {
         setQuery(query);
      }

      if (list != null)
      {
         for (int i = 0; i < list.length; i++)
         {
            // provide link
            ((BrowseColumnWidget) list[i].getWidget()).setElement(list[i]);
         }
      }
   }

   /**
    * Finish initialization of the browse widget. 
    */
   public void finishSetup()
   {
      if (initDone)
         return;
    
      if (query == null)
      {
         initDone = true;
         return;
      }
      
      // finish initialization of the widget
      try
      {
         // retrieve the first record only if the query is already opened
         if (query.isOpen().booleanValue() && query.currentRowImpl().isUnknown())
            query.first();
      }
      catch (QueryOffEndException exc)
      {
         // OK to ignore;  indicates query has no results and browse will be
         // empty initially.
      }
      config.queryId = getId();
      pushScreenDefinition();
      
      initDone = true;
   }

   /**
    * Invoked when the query with which this browse widget is associated is
    * explicitly repositioned.
    */
   public void queryRepositioned()
   {
      if (ignoreRepo > 0)
      {
         return;
      }

      LogicalTerminal.refreshScrollRow(this, false, false);
   }

   /**
    * Force refreshing of the browse at client side.
    */
   public void forceRefresh()
   {
      if (ignoreRepo > 0)
      {
         return;
      }

      LogicalTerminal.refreshScrollRow(this, false, false);
   }
   
   /**
    * Invoked when the query with which this browse widget is associated is
    * explicitly closed.
    */
   public void queryClosed()
   {
      LogicalTerminal.queryClosed(this);
      maximumKnownRowIndex = null;
   }

   /**
    * Determines if the specified browse row is "available", i.e. it is not deleted.
    *
    * @param  rowIndex
    *         0-base index of the target row in the browse model.
    *
    * @return <code>true</code> if the row is "available", i.e. it is not deleted.
    */
   public boolean isRowAvailable(int rowIndex)
   {
      BaseDataType[][] data = getRows(rowIndex, 1, true);
      return data.length > 0 && rowIndex == ((integer) data[0][0]).intValue();
   }

   /**
    * Compute the expression kept in the {@link #dynTitle} field and set it
    * as the current browse title. This operation will cancel the old
    * {@link #dynTitle} by setting it to null.
    */
   void updateDynamicTitle()
   {
      if (dynTitle != null)
      {
         character title = dynTitle.execute();
         String    txt   = (title.isUnknown() ? "" : title.toStringMessage());

         setTitle(txt);
      }
   }

   /**
    * Delete the row with the specified index. Query remains repositioned to the next row after
    * the deleted one.
    *
    * @param  rowIndex
    *         0-based index of the deleted row.
    * @param  refreshBrowse
    *         If <code>true</code> then the browse is refreshed.
    *
    * @return <code>true</code> on success.
    */
   private logical deleteRow(int rowIndex, boolean refreshBrowse)
   {
      reposition(new int64(rowIndex + 1), false);
      logical res = deleteResultListEntry(true, false);
      if (!res.booleanValue())
         return res;

      if (refreshBrowse)
         LogicalTerminal.refreshAfterDeletion(this, rowIndex);
      return new logical(true);
   }

   /**
    * Common reposition code.
    * 
    * @param   row
    *          1-based row to set.
    * @param   fetch
    *          <code>true</code> to auto-fetch the record after reposition,
    *          <code>false</code> to move the cursor only.
    */
   private void reposition(int64 row, boolean fetch)
   {
      if (query == null)
      {
         return;
      }

      boolean fOr = query.isFetchOnReposition(); 
      
      try
      {
         ignoreRepo++;
         
         if (!fetch)
            query.setFetchOnReposition(false);
         
         query.reposition(row);
      }
      finally
      {
         ignoreRepo--;
         
         if (!fetch)
            query.setFetchOnReposition(fOr);
      }
   }

   /**
    * Check if the underlying query is open and the result set contains records. If it doesn't
    * have any records, error message "Browse widget must contain at least one record before
    * <i>methodName</i> method is used" is displayed.
    *
    * @param   methodName
    *          The name of 4GL method that calls this function. Used in the error text.
    *
    * @return  <code>true</code> if the query is open and the result set contains at least one
    *          record.
    */
   private boolean checkNoRecords(String methodName)
   {
      if (query == null ||
          !query.isOpen().getValue() ||
          query.getNumResults().isUnknown() ||
          query.getNumResults().intValue() == 0)
      {
         final String msg = String.format("Browse widget must contain at least one " +
               "record before %s method is used", methodName);
         ErrorManager.recordOrShowError(2108, msg, false, false);
         return false;
      }

      return true;
   }

   /**
    * Get number of rows in the browse viewport set either by DOWN or SIZE options.
    *
    * @return  number of rows in the browse viewport.
    */
   private int getBrowsePageRowCount()
   {
      if (pageRowsCount == null)
         pageRowsCount = LogicalTerminal.getBrowsePageRowCount(this);

      return pageRowsCount;
   }

   /**
    * Displays the "Invalid row index specified for browse method <i>methodName</i>" error
    * message.
    *
    * @param methodName
    *        The name of 4GL method that calls this function. Used in the error text.
    */
   private void errorInvalidRowIndex(String methodName)
   {
      final String msg = "Invalid row index specified for browse method " + methodName;
      ErrorManager.recordOrShowError(new int[] { 385 },
                                     new String[] { msg },
                                     false,
                                     false,
                                     true,
                                     false,
                                     false);
   }

   /**
    * Displays the "Invalid use of browse method <i>methodName</i>. There are no selected rows."
    * error message.
    *
    * @param methodName
    *        The name of 4GL method that calls this function. Used in the error text.
    */
   private void errorNoSelectedRows(String methodName)
   {
      final String msg = String.format(
            "Invalid use of browse method %s. There are no selected rows", methodName);
      ErrorManager.recordOrShowError(new int[] { 382 },
                                     new String[] { msg },
                                     false,
                                     false,
                                     true,
                                     false,
                                     false);
   }

   /**
    * Performs full cycle of DELETE-CURRENT-ROW or DELETE-SELECTED-ROW processing.
    *
    * @param  selIndex
    *         0-based position of the row in the list of selected rows. <code>-1</code> for the
    *         last selected row.
    * @param  methodName
    *         4GL method name, used for error messages.
    *
    * @return <code>true</code> of success.
    */
   private logical deleteSingleRow(int selIndex, String methodName)
   {
      if (!checkNoRecords(methodName))
      {
         return new logical(false);
      }

      logical res = new logical(false);
      try
      {
         int targetRow = LogicalTerminal.getSelectedRowIndex(this, selIndex);
         res = deleteRow(targetRow, true);
      }
      catch (ErrorConditionException e)
      {
         handleError(e, methodName);
      }

      return res;
   }

   /**
    * Displays the appropriate error message for an exception with specific 4GL error code.
    *
    * @param  exc
    *         Exception with specific 4GL error code.
    *
    * @param  methodName
    *         4GL method name, used in the error message.
    */
   private void handleError(ErrorConditionException exc, String methodName)
   {
      int errorCode = exc.getProgressErrorCode();
      if (errorCode == 385)
      {
         errorInvalidRowIndex(methodName);
      }
      else if (errorCode == 382)
      {
         errorNoSelectedRows(methodName);
      }
      else
      {
         throw new IllegalStateException("Unknown error code: " + errorCode);
      }
   }

   /**
    * If necessary, init MAX-DATA-GUESS after the query has been (re)opened.
    */
   private void checkMaxDataGuess()
   {
      if (maximumKnownRowIndex != null)
         return;

      integer size = query.size();
      maximumKnownRowIndex = size.isUnknown() ? 0 : size.intValue();
      if (query.isNativelyPreselect())
      {
         setMaxDataGuess(maximumKnownRowIndex);
         realDataGuessUsed = true;
      }
      else if (maximumKnownRowIndex > DATA_GUESS_DEFAULT_VALUE)
      {
         setMaxDataGuess(maximumKnownRowIndex);
         realDataGuessUsed = true;
      }
      else
      {
         setMaxDataGuess(DATA_GUESS_DEFAULT_VALUE);
         realDataGuessUsed = false;
      }
   }
}
