/*
** Module   : ClientBrowseModel.java
** Abstract : client side adapter for the BrowseModel interface
**
** Copyright (c) 2005-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description------------------------------
** 001 SIY 2005xxxx   @23595 Created initial version.
** 002 SIY 20060315   @25126 Changes for complete BROWSE implementation.
** 003 GES 20060808   @28495 Use int instead of long for the backing methods that support the 
**                           server exports.
** 004 SIY 20061105   @30981 Added getCurrentRow() method.
** 005 SIY 20070504   @33423 Updated getRowCount().
** 006 SIY 20111101          Moved from CHUI package. Minor cleanups/refactorings in preparation 
**                           to GUI client.
** 007 CA  20140805          Explicitly import any classes from chui package(s).
** 008 SVL 20150212          Modified getRows signature.
** 009 SVL 20150411          Added isRowAvailable.
*/

package com.goldencode.p2j.ui.client;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.util.*;

/**
 * This class implements client side of the <code>BrowseModel</code> used by
 * the <code>Browse</code> widget. The implementation is just forwards all
 * requests to the remote side using ServerExports interface.
 * 
 * @author  SIY
 */
public class ClientBrowseModel
implements BrowseModel
{
   /** ID of the query at remote (server) side. */
   private int queryId = -1;
   
   /** Cache collected columns. */
   private BrowseColumnConfig[] columnsCache = null;
   
   /**
    * Construct an instance of adapter class using specified query ID and
    * reference to server interface.
    * 
    * @param   queryId
    *          ID of the query to forward request to.
    */
   public ClientBrowseModel(int queryId)
   {
      this.queryId = queryId;
   }
   
   /**
    * Return information about columns.
    * 
    * @return  A meta information about columns.
    */
   @Override
   public BrowseColumnConfig[] getColumns()
   {
      if (columnsCache == null)
         columnsCache = ThinClient.getInstance().getColumns(queryId);
      
      return columnsCache;
   }

   /**
    * Get number of available rows starting from the given one.
    *
    * @param   startRow
    *          Initial row to start count rows from.
    *
    * @return  Number of available rows.
    */
   @Override
   public int getRowCount(int startRow)
   {
      return ThinClient.getInstance().getRowCount(queryId, startRow);
   }

   /**
    * Retrieves several subsequent rows from the database.
    *
    * @param    startRowIndex
    *           Index of the first row to retrieve.
    * @param    count
    *           Maximum count of rows to return.
    * @param    forward
    *           <code>true</code> if rows are retrieved in the forward direction,
    *           <code>false</code> if rows are retrieved in the backward direction,
    *
    * @return   Array of rows. Each row is array of cell values, but the first element of the row
    *           array is 0-based index of the row in the browse model.
    *           Note that if rows are fetched in the backward direction, the resulting rows are
    *           returned in the straight order starting from the row with the least row number.
    *           If no query specified, an array of size 0 is returned.
    */
   @Override
   public BaseDataType[][] getRows(int startRowIndex, int count, boolean forward)
   {
      return ThinClient.getInstance().getRows(queryId, startRowIndex, count, forward);
   }

   /**
    * Updates selected cells in the given row.
    * 
    * @param   rowIndex
    *          Index of the row to update.
    * @param   cellIndices
    *          Cells for update indexes.
    * @param   cellValues
    *          Cells for update new values.
    * 
    * @throws  DataValidationException
    *          New values don't passed database validation rules.
    */
   @Override
   public void updateRow(int            rowIndex,
                         int[]          cellIndices,
                         BaseDataType[] cellValues)
   throws DataValidationException
   {
      ThinClient.getInstance().updateRow(queryId,
                                         rowIndex,
                                         cellIndices,
                                         cellValues);
   }

   /**
    * Returns <code>true</code> if specified row is new and is not present
    * in the data base.
    * 
    * @param   rowIndex
    *          Index of the row to return information about.
    * @return  <code>true</code> if row is new and <code>false</code>
    *          otherwise.
    */
   @Override
   public boolean isNewRow(int rowIndex)
   {
      return ThinClient.getInstance().isNewRow(queryId, rowIndex);
   }

   /**
    * Inform server about current row in the browse.
    * 
    * @param   rowIndex
    *          0-based index of the current row.
    *          
    * @return  New current row.
    */
   @Override
   public int setCurrentRow(int rowIndex)
   {
      return ThinClient.getInstance().setCurrentRow(queryId, rowIndex);
   }

   /**
    * Get current row in the model.
    * 
    * @return  The current row index in the model.
    */
   @Override
   public int getCurrentRow()
   {
      return ThinClient.getInstance().getCurrentRow(queryId);
   }

   /**
    * Determines if the specified browse row is "available", i.e. it is not deleted.
    *
    * @param  rowIndex
    *         0-base index of the target row in the browse model.
    *
    * @return <code>true</code> if the row is "available", i.e. it is not deleted.
    */
   @Override
   public boolean isRowAvailable(int rowIndex)
   {
      return ThinClient.getInstance().isRowAvailable(queryId, rowIndex);
   }
}
